"""
Code for `sarif summary` command.
"""

import os
from typing import List

from sarif.sarif_file import SarifFileSet


def generate_summary(
    input_files: SarifFileSet, output: str, output_multiple_files: bool
):
    """
    Generate a summary of the issues from the SARIF files.
    sarif_dict is a dict from filename to deserialized SARIF data.
    output_file is the name of a text file to write, or if None, the summary is written to the
    console.
    """
    output_file = output
    if output_multiple_files:
        for input_file in input_files:
            output_file_name = (
                input_file.get_file_name_without_extension() + "_summary.txt"
            )
            output_file = os.path.join(output, output_file_name)
            summary_lines = _generate_summary(input_file)
            print(
                "Writing summary of",
                input_file.get_file_name(),
                "to",
                output_file_name,
            )
            with open(output_file, "w", encoding="utf-8") as file_out:
                file_out.writelines(line + "\n" for line in summary_lines)
        output_file_name = "static_analysis_summary.txt"
        output_file = os.path.join(output, output_file_name)

    summary_lines = _generate_summary(input_files)
    if output:
        print(
            "Writing summary of",
            input_files.get_description(),
            "to",
            output_file,
        )
        with open(output_file, "w", encoding="utf-8") as file_out:
            file_out.writelines(line + "\n" for line in summary_lines)
    else:
        for lstr in summary_lines:
            print(lstr)


def _generate_summary(input_files: SarifFileSet) -> List[str]:
    """
    For each severity level (in priority order): create a list of the errors of
    that severity, print out how many there are and then do some further analysis
    of which error codes are present.
    """
    ret = []
    report = input_files.get_report()
    for severity in report.get_severities():
        result_count = report.get_issue_count_for_severity(severity)
        issue_type_histogram = report.get_issue_type_histogram_for_severity(severity)
        ret.append(f"\n{severity}: {result_count}")
        ret += [f" - {key}: {count}" for (key, count) in issue_type_histogram.items()]
    filter_stats = input_files.get_filter_stats()
    if filter_stats:
        ret.append(f"\nResults were filtered by {filter_stats}")
    return ret



"""
Code to load SARIF files from disk.
"""

import glob
import json
import os

from sarif.sarif_file import has_sarif_file_extension, SarifFile, SarifFileSet


def _add_path_to_sarif_file_set(path, sarif_file_set):
    if os.path.isdir(path):
        sarif_file_set.add_dir(_load_dir(path))
        return True
    if os.path.isfile(path):
        sarif_file_set.add_file(load_sarif_file(path))
        return True
    return False


def load_sarif_files(*args) -> SarifFileSet:
    """
    Load SARIF files specified as individual filenames or directories.  Return a SarifFileSet
    object.
    """
    ret = SarifFileSet()
    if args:
        for path in args:
            path_exists = _add_path_to_sarif_file_set(path, ret)
            if not path_exists:
                for resolved_path in glob.glob(path, recursive=True):
                    if _add_path_to_sarif_file_set(resolved_path, ret):
                        path_exists = True
            if not path_exists:
                print(f"Warning: input path {path} not found")
    return ret


def _load_dir(path):
    subdir = SarifFileSet()
    for dirpath, _dirnames, filenames in os.walk(path):
        for filename in filenames:
            if has_sarif_file_extension(filename):
                subdir.add_file(load_sarif_file(os.path.join(dirpath, filename)))
    return subdir


def load_sarif_file(file_path: str) -> SarifFile:
    """
    Load JSON data from a file and return as a SarifFile object.
    As per https://tools.ietf.org/id/draft-ietf-json-rfc4627bis-09.html#rfc.section.8.1, JSON
    data SHALL be encoded in utf-8.
    """
    try:
        with open(file_path, encoding="utf-8-sig") as file_in:
            data = json.load(file_in)
        return SarifFile(file_path, data)
    except Exception as exception:
        raise IOError(f"Cannot load {file_path}") from exception



"""
Code for `sarif emacs` command.
"""

from datetime import datetime
import os

from jinja2 import Environment, FileSystemLoader, select_autoescape

from sarif import sarif_file

_THIS_MODULE_PATH = os.path.dirname(__file__)

_TEMPLATES_PATH = os.path.join(_THIS_MODULE_PATH, "templates")

_ENV = Environment(
    loader=FileSystemLoader(searchpath=_TEMPLATES_PATH),
    autoescape=select_autoescape(),
)


def generate_compile(
    input_files: sarif_file.SarifFileSet,
    output: str,
    output_multiple_files: bool,
    date_val: datetime = datetime.now(),
):
    """
    Generate txt file from the input files.
    """
    output_file = output
    if output_multiple_files:
        for input_file in input_files:
            output_file_name = input_file.get_file_name_without_extension() + ".txt"
            print(
                "Writing results for",
                input_file.get_file_name(),
                "to",
                output_file_name,
            )
            _generate_single_txt(
                input_file, os.path.join(output, output_file_name), date_val
            )
        output_file = os.path.join(output, ".compile.txt")
    source_description = input_files.get_description()
    print(
        "Writing results for",
        source_description,
        "to",
        os.path.basename(output_file),
    )
    _generate_single_txt(input_files, output_file, date_val)


def _generate_single_txt(input_file, output_file, date_val):
    all_tools = input_file.get_distinct_tool_names()
    report = input_file.get_report()

    total_distinct_issue_codes = 0
    problems = []
    severities = report.get_severities()

    for severity in severities:
        distinct_issue_codes = report.get_issue_type_count_for_severity(severity)

        total_distinct_issue_codes += distinct_issue_codes

        severity_details = _enrich_details(
            report.get_issues_grouped_by_type_for_severity(severity)
        )

        severity_section = {
            "type": severity,
            "count": distinct_issue_codes,
            "details": severity_details,
        }

        problems.append(severity_section)

    filtered = None
    filter_stats = input_file.get_filter_stats()
    if filter_stats:
        filtered = f"Results were filtered by {filter_stats}."

    template = _ENV.get_template("sarif_emacs.txt")
    txt_content = template.render(
        report_type=", ".join(all_tools),
        report_date=date_val,
        severities=", ".join(severities),
        total=total_distinct_issue_codes,
        problems=problems,
        filtered=filtered,
    )

    with open(output_file, "wt", encoding="utf-8") as file_out:
        file_out.write(txt_content)


def _enrich_details(records_of_severity):
    return [
        {"code": key, "count": len(records), "details": records}
        for (key, records) in records_of_severity.items()
    ]



"""
Code for `sarif codeclimate` command.
"""

import os
import json
import hashlib

from sarif.sarif_file import SarifFileSet

_SEVERITIES = {"none": "info", "note": "info", "warning": "minor", "error": "major"}


def generate(input_files: SarifFileSet, output: str, output_multiple_files: bool):
    """
    Generate a JSON file in Code Climate schema containing the list of issues from the SARIF files.
    See https://github.com/codeclimate/platform/blob/master/spec/analyzers/SPEC.md
    Gitlab usage guide - https://docs.gitlab.com/ee/ci/testing/code_quality.html#implement-a-custom-tool
    """
    output_file = output
    if output_multiple_files:
        for input_file in input_files:
            output_file_name = input_file.get_file_name_without_extension() + ".json"
            print(
                "Writing Code Climate JSON summary of",
                input_file.get_file_name(),
                "to",
                output_file_name,
            )
            _write_to_json(
                input_file.get_records(), os.path.join(output, output_file_name)
            )
            filter_stats = input_file.get_filter_stats()
            if filter_stats:
                print(f"  Results are filtered by {filter_stats}")
        output_file = os.path.join(output, "static_analysis_output.json")
    source_description = input_files.get_description()
    print(
        "Writing Code Climate JSON summary for",
        source_description,
        "to",
        os.path.basename(output_file),
    )
    _write_to_json(input_files.get_records(), output_file)
    filter_stats = input_files.get_filter_stats()
    if filter_stats:
        print(f"  Results are filtered by {filter_stats}")


def _write_to_json(list_of_errors, output_file):
    """
    Write out the errors to a JSON file according to Code Climate specification.
    """
    content = []
    for record in list_of_errors:
        severity = _SEVERITIES.get(record.get("Severity", "warning"), "minor")

        # split Code value to extract error ID and description
        rule = record["Code"]
        description = record["Description"]

        path = record["Location"]
        line = record["Line"]

        fingerprint = hashlib.md5(
            f"{description} {path} ${line}`]".encode()
        ).hexdigest()

        # "categories" property is not used in GitLab but marked as "required" in Code Climate spec.
        # There is no easy way to determine a category so the fixed value is set.
        content.append(
            {
                "type": "issue",
                "check_name": rule,
                "description": description,
                "categories": ["Bug Risk"],
                "location": {"path": path, "lines": {"begin": line}},
                "severity": severity,
                "fingerprint": fingerprint,
            }
        )

    with open(output_file, "w", encoding="utf-8") as file_out:
        json.dump(content, file_out, indent=4)



"""
Code for `sarif info` command.
"""

import datetime
import os

from sarif.sarif_file import SarifFileSet

_BYTES_PER_MIB = 1024 * 1024
_BYTES_PER_KIB = 1024


def _property_bag_stats(object_list):
    tally = {}
    universal_property_keys = []
    partial_properties = []
    if object_list:
        for obj in object_list:
            for key in obj.get("properties", {}):
                tally[key] = tally[key] + 1 if key in tally else 1
        object_count = len(object_list)
        universal_property_keys = [
            key for (key, count) in tally.items() if count == object_count
        ]

        def tally_rank(key_count_pair):
            # Sort by descending tally then alphabetically
            return (-key_count_pair[1], key_count_pair[0])

        partial_properties = [
            {"key": key, "count": count, "percent": 100 * count / object_count}
            for (key, count) in sorted(tally.items(), key=tally_rank)
            if count < object_count
        ]
    return universal_property_keys, partial_properties


def _generate_info_to_file(sarif_files, file_out):
    file_count = False
    for input_file in sarif_files:
        file_count += 1
        file_path = input_file.get_abs_file_path()
        file_stat = os.stat(file_path)
        size_in_bytes = file_stat.st_size
        if size_in_bytes > _BYTES_PER_MIB:
            readable_size = f"{file_stat.st_size / _BYTES_PER_MIB:.1f} MiB"
        else:
            readable_size = (
                f"{(file_stat.st_size + _BYTES_PER_KIB - 1) // _BYTES_PER_KIB} KiB"
            )
        print(input_file.get_abs_file_path(), file=file_out)
        print(f"  {file_stat.st_size} bytes ({readable_size})", file=file_out)
        print(
            f"  modified: {datetime.datetime.fromtimestamp(file_stat.st_mtime)}, "
            f"accessed: {datetime.datetime.fromtimestamp(file_stat.st_atime)}, "
            f"ctime: {datetime.datetime.fromtimestamp(file_stat.st_ctime)}",
            file=file_out,
        )
        run_count = len(input_file.runs)
        print(f"  {run_count} runs" if run_count != 1 else "  1 run", file=file_out)
        for run_index, run in enumerate(input_file.runs):
            if run_count != 1:
                print(f"  Run #{run_index + 1}:", file=file_out)
            print(f"    Tool: {run.get_tool_name()}", file=file_out)
            conversion_tool = run.get_conversion_tool_name()
            if conversion_tool:
                print(f"    Conversion tool: {conversion_tool}", file=file_out)
            results = run.get_results()
            result_count = len(results)
            print(
                f"    {result_count} results" if result_count != 1 else "    1 result",
                file=file_out,
            )
            universal_property_keys, partial_properties = _property_bag_stats(results)
            ppk_string = (
                ", ".join(
                    "{} {}/{} ({:.1f} %)".format(
                        p["key"], p["count"], result_count, p["percent"]
                    )
                    for p in partial_properties
                )
                if partial_properties
                else None
            )
            if universal_property_keys:
                upk_string = ", ".join(universal_property_keys)
                if partial_properties:
                    print(
                        f"    Result properties: all results have properties: {upk_string}; "
                        f"some results have properties: {ppk_string}",
                        file=file_out,
                    )
                else:
                    print(
                        f"    All results have properties: {upk_string}",
                        file=file_out,
                    )
            elif partial_properties:
                print(
                    f"    Result properties: {ppk_string}",
                    file=file_out,
                )
        print(file=file_out)
    return file_count


def generate_info(sarif_files: SarifFileSet, output: str):
    """
    Print structure information about the provided `sarif_files`.
    """
    if output:
        with open(output, "w", encoding="utf-8") as file_out:
            file_count = _generate_info_to_file(sarif_files, file_out)
        if file_count:
            files_string = (
                "1 SARIF file" if file_count == 1 else f"{file_count} SARIF files"
            )
            print("Wrote information about", files_string, "to", output)
    else:
        file_count = _generate_info_to_file(sarif_files, None)
    if file_count == 0:
        print(
            "No SARIF files found.  Try passing a path of a SARIF file or containing SARIF files."
        )



"""
Generate summary of given JSON or given JSON Document in to Microsoft Word Documents.
This functionality uses a python module called

python-docx - a Python library for creating and updating Microsoft Word (.docx) files.

https://python-docx.readthedocs.io/

"""

from datetime import datetime
import os
from typing import Union

import docx
from docx import oxml
from docx import shared
from docx.enum import text
from docx.oxml import ns

from sarif import charts, sarif_file
from sarif.sarif_file_utils import combine_record_code_and_description


def generate_word_docs_from_sarif_inputs(
    input_files: sarif_file.SarifFileSet,
    image_file: Union[str, None],
    output: str,
    output_multiple_files: bool,
    date_val: datetime = datetime.now(),
):
    """
    Convert SARIF input to Word file output.
    """
    if not input_files:
        raise ValueError("No input files specified!")

    output_file = output
    output_file_name = output
    if output_multiple_files:
        for input_file in input_files:
            output_file_name = input_file.get_file_name_without_extension() + ".docx"
            print(
                "Writing Word summary of",
                input_file.get_file_name(),
                "to",
                output_file_name,
            )
            report = input_file.get_report()
            _generate_word_summary(
                input_file,
                report,
                os.path.join(output, output_file_name),
                image_file,
                date_val,
            )
        output_file_name = "static_analysis_output.docx"
        output_file = os.path.join(output, output_file_name)

    source_description = input_files.get_description()
    print("Writing Word summary of", source_description, "to", output_file_name)
    report = input_files.get_report()
    _generate_word_summary(input_files, report, output_file, image_file, date_val)


def _generate_word_summary(
    sarif_data, report, output_file, image_file: Union[str, None], date_val: datetime
):
    # Create a new document
    document = docx.Document()

    severities = report.get_severities()
    _add_heading_and_highlevel_info(
        document, sarif_data, report, severities, output_file, image_file, date_val
    )
    _dump_errors_summary_by_sev(document, report, severities)
    _dump_each_error_in_detail(document, report, severities)

    # finally, save the document.
    document.save(output_file)


def _add_heading_and_highlevel_info(
    document,
    sarif_data,
    report,
    severities,
    output_file,
    image_path: Union[str, None],
    date_val: datetime,
):
    tool_name = ", ".join(sarif_data.get_distinct_tool_names())
    heading = f"Sarif Summary: {tool_name}"

    if image_path:
        document.add_picture(image_path)
        last_paragraph = document.paragraphs[-1]
        last_paragraph.alignment = text.WD_PARAGRAPH_ALIGNMENT.CENTER

    document.add_heading(heading, 0)
    document.add_paragraph(f"Document generated on: {date_val}")

    sevs = ", ".join(severities)
    document.add_paragraph(
        f"Total number of various severities ({sevs}): {sarif_data.get_result_count()}"
    )
    filter_stats = sarif_data.get_filter_stats()
    if filter_stats:
        document.add_paragraph(f"Results were filtered by {filter_stats}.")

    pie_chart_image_file_path = output_file.replace(".docx", "_severity_pie_chart.png")
    if charts.generate_severity_pie_chart(report, pie_chart_image_file_path):
        document.add_picture(pie_chart_image_file_path)
    last_paragraph = document.paragraphs[-1]
    last_paragraph.alignment = text.WD_PARAGRAPH_ALIGNMENT.CENTER

    document.add_page_break()


def _dump_errors_summary_by_sev(document, report, severities):
    """
    For each severity level (in priority order): create a list of the errors of
    that severity, print out how many there are and then do some further analysis
    of which error codes are present.
    """
    for severity in severities:
        errors_of_severity = report.get_issue_type_count_for_severity(severity)
        document.add_heading(f"Severity : {severity} [ {errors_of_severity} ]", level=1)
        sorted_dict = report.get_issue_type_histogram_for_severity(severity)
        if sorted_dict:
            for key, count in sorted_dict.items():
                document.add_paragraph(f"{key}: {count}", style="List Bullet")
        else:
            document.add_paragraph("None", style="List Bullet")


def _dump_each_error_in_detail(document, report, severities):
    """
    Write out the errors to a table so that a human can do further analysis.
    """
    document.add_page_break()

    for severity in severities:
        errors_of_severity = report.get_issues_for_severity(severity)
        # Sample:
        # [{'Location': 'C:\\Max\\AccessionAndroid\\scripts\\parse_coverage.py', 'Line': 119,
        #       'Severity': 'error', 'Code': 'DS126186 Disabled certificate validation'},
        # {'Location': 'C:\\Max\\AccessionAndroid\\scripts\\parse_code_stats.py', 'Line': 61,
        #       'Severity': 'error', 'Code': 'DS126186 Disabled certificate validation'},
        # ]
        if errors_of_severity:
            document.add_heading(f"Severity : {severity}", level=2)
            table = document.add_table(rows=1 + len(errors_of_severity), cols=3)

            table.style = "Table Grid"  # ColorfulGrid-Accent5'
            table.autofit = False

            table.alignment = text.WD_TAB_ALIGNMENT.CENTER

            # Cell widths
            widths = [shared.Inches(2), shared.Inches(4), shared.Inches(0.5)]

            # To avoid performance problems with large tables, prepare the entries first in this
            # list, then iterate the table cells and copy them in.
            # First populate the header row
            cells_text = ["Code", "Location", "Line"]

            hdr_cells = table.rows[0].cells
            for i in range(3):
                table.rows[0].cells[i]._tc.get_or_add_tcPr().append(
                    oxml.parse_xml(
                        r'<w:shd {} w:fill="5fe3d8"/>'.format(ns.nsdecls("w"))
                    )
                )
                run = hdr_cells[i].paragraphs[0].add_run(cells_text[i])
                run.bold = True
                hdr_cells[i].paragraphs[
                    0
                ].alignment = text.WD_PARAGRAPH_ALIGNMENT.CENTER
                hdr_cells[i].width = widths[i]

            for eachrow in errors_of_severity:
                cells_text += [
                    combine_record_code_and_description(eachrow),
                    eachrow["Location"],
                    str(eachrow["Line"]),
                ]

            # Note: using private property table._cells to avoid performance issue.  See
            # https://stackoverflow.com/a/69105798/316578
            col_index = 0
            for cell, cell_text in zip(table._cells, cells_text):
                cell.text = cell_text
                cell.width = widths[col_index]
                col_index = col_index + 1 if col_index < 2 else 0
        else:
            document.add_heading(f"Severity : {severity}", level=2)
            document.add_paragraph("None", style="List Bullet")



"""
Reusable utility functions for handling the SARIF format.

Primarily interrogating the `result` JSON defined at
https://docs.oasis-open.org/sarif/sarif/v2.1.0/cs01/sarif-v2.1.0-cs01.html#_Toc16012594
"""

import textwrap
from typing import Literal, Tuple, Union

# SARIF severity levels as per
# https://docs.oasis-open.org/sarif/sarif/v2.1.0/sarif-v2.1.0.html#_Toc141790898
SARIF_SEVERITIES_WITHOUT_NONE = ["error", "warning", "note"]
SARIF_SEVERITIES_WITH_NONE = SARIF_SEVERITIES_WITHOUT_NONE + ["none"]


def combine_code_and_description(code: str, description: str) -> str:
    """
    Combine code and description into one string, keeping total length under 120 characters.
    """
    length_budget = 120
    if code:
        code = code.strip()
        length_budget -= len(code) + 1  # Allow issue code and space character
    continuation_placeholder = " ..."
    # Allow extra space when truncating for continuation characters
    length_budget_pre_continuation = length_budget - len(continuation_placeholder)
    if length_budget_pre_continuation < 10:
        # Don't include description if it would be very short due to long code
        return code
    if description:
        if "\n" in description:
            description = description[: description.index("\n")]
        if description.startswith(code):
            # Don't duplicate the code
            description = description[len(code) :]
        description = description.strip()
    if description:
        if len(description) > length_budget:
            shorter_description = textwrap.shorten(
                description,
                width=length_budget_pre_continuation,
                placeholder=continuation_placeholder,
            )
            if len(shorter_description) < length_budget_pre_continuation - 40:
                # Word wrap shortens the description significantly, so truncate mid-word instead
                description = (
                    description[:length_budget_pre_continuation]
                    + continuation_placeholder
                )
            else:
                description = shorter_description
        if code:
            return f"{code.strip()} {description}"
        return description
    if code:
        return code
    return "<NONE>"


def combine_record_code_and_description(record: dict) -> str:
    """
    Combine code and description fields into one string.
    """
    return combine_code_and_description(record["Code"], record["Description"])


def read_result_location(result) -> Tuple[str, str]:
    """
    Extract the file path and line number strings from the Result.

    Tools store this in different ways, so this function tries a few different JSON locations.
    """
    file_path = None
    line_number = None
    locations = result.get("locations", [])
    if locations and isinstance(locations, list):
        location = locations[0]
        physical_location = location.get("physicalLocation", {})
        # SpotBugs has some errors with no line number so deal with them by just leaving it at 1
        line_number = physical_location.get("region", {}).get("startLine", None)
        # For file name, first try the location written by DevSkim
        file_path = (
            location.get("physicalLocation", {})
            .get("address", {})
            .get("fullyQualifiedName", None)
        )
        if not file_path:
            # Next try the physical location written by MobSF and by SpotBugs (for some errors)
            file_path = (
                location.get("physicalLocation", {})
                .get("artifactLocation", {})
                .get("uri", None)
            )
        if not file_path:
            logical_locations = location.get("logicalLocations", None)
            if logical_locations:
                # Finally, try the logical location written by SpotBugs for some errors
                file_path = logical_locations[0].get("fullyQualifiedName", None)
    return (file_path, line_number)


def read_result_rule(result, run) -> Tuple[Union[dict, None], int]:
    """
    Returns the corresponding rule object for the specified result, plus its index
    in the rules array. Follows the rules at
    https://docs.oasis-open.org/sarif/sarif/v2.1.0/sarif-v2.1.0.html#_Toc141790895
    """
    ruleIndex = result.get("ruleIndex")
    ruleId = result.get("ruleId")
    rule = result.get("rule")

    if rule:
        if ruleIndex is None:
            ruleIndex = rule.get("index")

        if ruleId is None:
            ruleId = rule.get("id")

    rules = run.get("tool", {}).get("driver", {}).get("rules", [])

    if ruleIndex is not None and ruleIndex >= 0 and ruleIndex < len(rules):
        return (rules[ruleIndex], ruleIndex)

    if ruleId:
        for i, rule in enumerate(rules):
            if rule.get("id") == ruleId:
                return (rule, i)

    return (None, -1)


def read_result_invocation(result, run):
    """
    Extract the invocation metadata for the result, following the rules at
    https://docs.oasis-open.org/sarif/sarif/v2.1.0/sarif-v2.1.0.html#_Toc141790917
    """
    invocationIndex = result.get("provenance", {}).get("invocationIndex")
    if invocationIndex is None:
        return None

    invocations = run.get("invocations")

    if invocations and invocationIndex >= 0 and invocationIndex < len(invocations):
        return invocations[invocationIndex]

    return None


def read_result_severity(result, run) -> Literal["none", "note", "warning", "error"]:
    """
    Extract the severity level from the result following the rules at
    https://docs.oasis-open.org/sarif/sarif/v2.1.0/sarif-v2.1.0.html#_Toc141790898
    """
    severity = result.get("level")
    if severity:
        return severity

    # If kind has any value other than "fail", then if level is absent,
    # it SHALL default to "none"
    kind = result.get("kind", "fail")
    if kind and kind != "fail":
        return "none"

    # If kind has the value "fail" and level is absent, then...
    rule, ruleIndex = read_result_rule(result, run)
    if rule:
        # Honor the invocation's configuration override if present...
        invocation = read_result_invocation(result, run)
        if invocation:
            ruleConfigurationOverrides = invocation.get(
                "ruleConfigurationOverrides", []
            )
            override = next(
                (
                    override
                    for override in ruleConfigurationOverrides
                    if override.get("descriptor", {}).get("id") == rule.get("id")
                    or override.get("descriptor", {}).get("index") == ruleIndex
                ),
                None,
            )

            if override:
                overrideLevel = override.get("configuration", {}).get("level")
                if overrideLevel:
                    return overrideLevel

        # Otherwise, use the rule's default configuraiton if present...
        defaultConfiguration = rule.get("defaultConfiguration")
        if defaultConfiguration:
            severity = defaultConfiguration.get("level")
            if severity:
                return severity

    # Otherwise, fall back to warning
    return "warning"


def record_sort_key(record: dict) -> str:
    """Get a sort key for the record."""
    return (
        combine_record_code_and_description(record)
        + record["Location"]
        + str(record["Line"]).zfill(6)
    )
